package demo.spdeveloper.plugin.orgchart.service;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import sailpoint.api.IdIterator;
import sailpoint.api.ObjectUtil;
import sailpoint.api.SailPointContext;
import sailpoint.object.Filter;
import sailpoint.object.Identity;
import sailpoint.object.QueryOptions;
import sailpoint.tools.GeneralException;
import sailpoint.tools.Util;

public class IdentityService {
	
	private static final String ATT_MANAGES_COUNT = "managesCount";
	private static final String ATTR_INACTIVE = "inactive";
	private static final String ATT_MEMBERSHIP_COUNT = "membershipCount";
	private static final String ATTR_IDENTITY_ICON_IMG_ATTRIBUTE = "identityIconImgAttribute";
	private static final String ATTR_MEMBER_COUNT = "memberCount";
	private static final String ATTR_ATTRIBUTES = "attributes";
	private static final String ATTR_ID = "id";
	private static final String ATTR_NAME = "name";
	private static final String ATTR_DISPLAYNAME = "displayName";
	private static final String ATTR_TYPE = "type";
	private static final String ATTR_COLOR_CODE = "colorCode";
	private static final String ATTR_PARENT = "parentId";
	private static final String TYPE_WORKGROUP = "workgroup";
	private static final String TYPE_DEFAULT = "default";
	
	public static final Log logger = LogFactory.getLog(IdentityService.class);
	
	public static List<Map<String, Object>> getTreeNodes(SailPointContext context, String id) throws GeneralException {
		Identity identity = context.getObjectById(Identity.class, id);
		if (identity == null) {
			throw new GeneralException("Can not find identity object: " + id);
		}
		List<Map<String, Object>> nodes = new ArrayList<>();
		List<String> all = new ArrayList<>();
		List<String> childs = getChilds(context, id);
		logger.trace("childs: " + childs);
		all.addAll(childs);
		List<String> parents = getParents(context, identity, PluginSettingService.getSettingManagerLevels(), new ArrayList<>());
		String rootId = null;
		if (Util.nullSafeSize(parents) > 0) {
			rootId = parents.get(parents.size() - 1);
		} else {
			rootId = id;
		}
		logger.trace("parents: " + parents);
		all.addAll(parents);
		List<String> siblings = getSiblings(context, identity);
		logger.trace("siblings: " + siblings);
		all.addAll(siblings);
		Util.removeDuplicates(all);
		for (String _id : Util.safeIterable(all)) {
			Identity node;
			try {
				node = context.getObjectById(Identity.class, _id);
			} catch (GeneralException e) {
				logger.error("Failed to retrieve Identity object(" + _id + "): " + e);
				continue;
			}
			// Mandatory attributes: id, name, parentId, displayName
			try {
				nodes.add(getNodeMap(context, node, rootId));
			} catch (GeneralException e) {
				logger.error("Failed to build node map(" + _id + "): " + e);
				continue;
			}
		}
		if (logger.isTraceEnabled()) {
			logger.trace("Exit getTreeNodes, nodes: " + nodes);
		}
		return nodes;
	}
	
	private static List<String> getChilds(SailPointContext context, String id) 
			throws GeneralException {
		List<String> childs = new ArrayList<>();
		childs.addAll(getManagedMembers(context, id));
		childs.addAll(getOwnedWorkgroups(context, id));
		return childs;
	}
	
	private static List<String> getManagedMembers(SailPointContext context, String id) 
			throws GeneralException {
		List<String> members = new ArrayList<>();
		QueryOptions qo = new QueryOptions();
		Filter filter = Filter.eq("manager.id", id);
		qo.add(filter);
		Iterator<Object[]> it = context.search(Identity.class, qo, "id");
		IdIterator idIt = new IdIterator(context, it);
		while (idIt.hasNext()) {
			String _id = idIt.next();
			members.add(_id);
		}
		return members;
	}
	
	private static List<String> getOwnedWorkgroups(SailPointContext context, String id) 
			throws GeneralException {
		List<String> workgroups = new ArrayList<>();
		QueryOptions qo = new QueryOptions();
		Filter filter = Filter.and(Filter.eq("owner.id", id), Filter.eq("workgroup", true));
		qo.add(filter);
		Iterator<Object[]> it = context.search(Identity.class, qo, "id");
		IdIterator idIt = new IdIterator(context, it);
		while (idIt.hasNext()) {
			String _id = idIt.next();
			workgroups.add(_id);
		}
		return workgroups;
	}
	
	private static int getManagedCount(SailPointContext context, String id) throws GeneralException {
		QueryOptions qo = new QueryOptions();
		Filter filter1 = Filter.and(Filter.eq("manager.id", id), Filter.eq("workgroup", false));
		Filter filter2 = Filter.and(Filter.eq("owner.id", id), Filter.eq("workgroup", true));
		Filter filter = Filter.or(filter1, filter2);
		qo.add(filter);
		return context.countObjects(Identity.class, qo);
	}
	
	private static List<String> getParents(SailPointContext context, Identity identity, int level, List<String> reportTos) {
		logger.trace("getParents, level: " + level);
		logger.trace("getParents, identityName: " + identity.getName());
		logger.trace("getParents, reportTos: " + reportTos);
		if (level == 0) {
			return reportTos;
		} else {
			if (identity.getManager() != null) {
				String managerId = identity.getManager().getId();
				if (reportTos.contains(managerId)) {
					// Close loop found
					return reportTos;
				} else {
					reportTos.add(managerId);
					return getParents(context, identity.getManager(), --level, reportTos);
				}
			} else if (identity.isWorkgroup() && identity.getOwner() != null) { // Workgroup
				String ownerId = identity.getOwner().getId();
				if (reportTos.contains(ownerId)) {
					// Close loop found
					return reportTos;
				} else {
					reportTos.add(ownerId);
					return getParents(context, identity.getOwner(), --level, reportTos);
				}
			} else {
				return reportTos;
			}
		}
	}
	
	/**
	 * Return siblings including self
	 * 
	 * @param context
	 * @param identity
	 * @return
	 * @throws GeneralException
	 */
	private static List<String> getSiblings(SailPointContext context, Identity identity) 
			throws GeneralException {
		List<String> siblings = new ArrayList<>();
		Identity parent = null;
		if (identity.isWorkgroup()) {
			parent = identity.getOwner();
		} else {
			parent = identity.getManager();
		}		
		if (parent != null) {
			siblings = getChilds(context, parent.getId());
		} else { // No manager or owner
			siblings.add(identity.getId());
		}
		return siblings;
	}
	
	private static Map<String, Object> getNodeMap(SailPointContext context, Identity node, String rootId) 
			throws GeneralException {
		if (logger.isTraceEnabled()) {
			logger.trace("Enter getNodeMap...");
			logger.trace("Node: " + node.getName());
			logger.trace("Root Id: " + rootId);
		}
		Map<String, Object> nodeMap = new HashMap<>();
		nodeMap.put(ATTR_ID, node.getId());
		nodeMap.put(ATTR_NAME, node.getName());
		nodeMap.put(ATTR_DISPLAYNAME, node.getDisplayableName());
		if (node.isWorkgroup()) {
			nodeMap.putAll(getWorkgroupNodeMap(context, node, rootId));
		} else {
			nodeMap.putAll(getIdentityNodeMap(context, node, rootId));
		}
		
		if (logger.isTraceEnabled()) {
			logger.trace("Exit getNodeMap, nodeMap: " + nodeMap);
		}
		return nodeMap;
	}

	private static Map<String, Object> getWorkgroupNodeMap(SailPointContext context, Identity node, String rootId) 
			throws GeneralException {
		Map<String, Object> map = new HashMap<>();
		map.put(ATTR_TYPE, TYPE_WORKGROUP);
		map.put(ATTR_COLOR_CODE, getColorCode(TYPE_WORKGROUP));
		if (rootId != null && rootId.equals(node.getId())) {
			map.put(ATTR_PARENT, null);
		} else {
			map.put(ATTR_PARENT, getParentNodeId(node, node.getOwner()));
		}
		map.put(ATT_MANAGES_COUNT, getManagedCount(context, node.getId()));
		// Custom attributes
		List<String> customWorkgroupCardAttrs = PluginSettingService.getSettingWorkgroupCardAttrs();
		for (String attr : Util.safeIterable(customWorkgroupCardAttrs)) {
			if ("description".equals(attr)) {
				map.put(attr, node.getDescription());
			} else {
				map.put(attr, node.getAttribute(attr));
			}
		}
		map.put(ATTR_ATTRIBUTES, customWorkgroupCardAttrs);
		try {
			map.put(ATTR_MEMBER_COUNT, Util.nullSafeSize(getMembers(context, node, "id")));
		} catch (GeneralException e) {
			logger.error("Failed to get members for workgroup: " + node.getName());
		}
		return map;
	}

	private static Map<String, Object> getIdentityNodeMap(SailPointContext context, Identity node, String rootId) 
			throws GeneralException {
		Map<String, Object> map = new HashMap<>();
		map.put(ATTR_TYPE, (node.getType() != null ? node.getType() : "none"));
		map.put(ATTR_COLOR_CODE, getColorCode(node.getType()));
		if (rootId != null && rootId.equals(node.getId())) {
			map.put(ATTR_PARENT, null);
		} else {
			map.put(ATTR_PARENT, getParentNodeId(node, node.getManager()));
		}
		map.put(ATT_MANAGES_COUNT, getManagedCount(context, node.getId()));
		map.put(ATTR_INACTIVE, node.isInactive());
		// Custom attributes
		List<String> customIdentityCardAttrs = PluginSettingService.getSettingIdentityCardAttrs();
		for (String attr : Util.safeIterable(customIdentityCardAttrs)) {
			map.put(attr, node.getAttribute(attr));
		}
		map.put(ATTR_ATTRIBUTES, customIdentityCardAttrs);
		map.putAll(getIconImg(node));
		// Workgroups
		List<Identity> workgroups = node.getWorkgroups();
		map.put(ATT_MEMBERSHIP_COUNT, Util.nullSafeSize(workgroups));
		return map;
	}
	
	private static String getColorCode(String type) {
		Map<String, String> colorCodeMap = PluginSettingService.getColorCodeSettingMap();
		if (Util.isNotNullOrEmpty(type) && colorCodeMap.containsKey(type)) {
			return colorCodeMap.get(type);
		} else {
			return colorCodeMap.get(TYPE_DEFAULT);
		}
	}
	
	private static String getParentNodeId(Identity node, Identity parent) {
		if (parent != null) {
			return parent.getId();	
		} else {
			return null;
		}
	}
	
	private static Map<String, Object> getIconImg(Identity node) {
		Map<String, Object> map = new HashMap<>();
		String iconImgAttr = PluginSettingService.getIdentityIconImgAttirbute();
		if (Util.isNotNullOrEmpty(iconImgAttr)) {
			map.put(ATTR_IDENTITY_ICON_IMG_ATTRIBUTE, iconImgAttr);
			if (Util.isNotNullOrEmpty(node.getStringAttribute(iconImgAttr))) {
				map.put(iconImgAttr, node.getStringAttribute(iconImgAttr));
			}
		}
		return map;
	}
	
	private static List<String> getMembers(SailPointContext context, Identity workgroup, String attribute) throws GeneralException {
		List<String> result = new ArrayList<>();
		List<String> props = new ArrayList<>();
		props.add(attribute);
		Iterator<Object[]> it = ObjectUtil.getWorkgroupMembers(context, workgroup, props);
		while (it.hasNext()) {
			String value = Util.otos(it.next()[0]);
			result.add(value);
		}
		Util.flushIterator(it);
		Collections.sort(result);
		return result;
	}

}
